//////////////////////////////////////////////////////////////////
//
//	WaitingARQ.h
//	
//	Waiting ARQ is used in the following way.
//  An ARQ is received in the Gk where the alias is a virtual queue ( CallCenter ).
//  The request is linked to a waiting queue ( WaitingARQlist ) while a GFA message  ( Get First Available Agent )
//  is sent via the status channel to a CallCenter queue system. An agent is selected and the call is presented to this agent. 
//  The call is answered by the agent ( pickup call) , a message GFC is sent back to the Gk with information 
//  callId , ip-address , alias.
//	The waiting ARQ list is scanned for this callId and the onARQ is called with a registred endpoint (the agent endpoint)  
//
// This work is published under the GNU Public License (GPL)
// see file COPYING for details.
// We also explicitely grant the right to link this code
// with the OpenH323 library.
//
// History:
// 	020300	initial version (Dennis Lazreg )
//
//////////////////////////////////////////////////////////////////


#ifndef _WaitingARQ_h__
#define _WaitingARQ_h__

#if HAS_WAITARQ

#include <ptlib.h>
#include <ptlib/sockets.h>
#include "h225.h" 
#include "h323.h"
#include "singleton.h"
#include "stl_supp.h"
#include "RasSrv.h"

#include <set>
#include <list>
#include <vector>
#include <string>


#define CTI_DEFAULT_TIME_OUT		10				// Time seconds 
#define CTI_VIRTUAL_QUEUE			"VirtualQueue"  // Name of the CTI virtual queue


class WaitingARQlist : public Singleton<WaitingARQlist> {
		
		class WaitingARQ {

			private:
				
			int									  m_requestSeqNumber;		// Request Unique number
  			H225_AdmissionRequest mm_ARQ;						    // Admission Request Record
			PTime									m_enterTime;				  // Enter time in this queue.
			PString								m_vAgent ;					  // Virtual Agent name..
  			endptr								m_reqEP;


			public:

				PString               m_epId;							// Endpoint Identifier
				H225_RasMessage       m_ARQ;						    // Admission Request Record
				PIPSocket::Address    m_rx_addr;					    // Sender IP address
				int                   m_callReference;     // Call Reference number  

				int DoARQ (void) const;
				int DoARJ (void) const;

				bool CompReqId (int req_ID) const;
				bool CompCallRef( int callRef ) const; 
				bool CompEndpointId (PString epId ) const;  
				bool IsTimeOut (int) const;
				
				PString getEndPointIdentifier () const;
				PString getAgentQueue() const ; 
				
				WaitingARQ(int seqNum, const H225_RasMessage & obj_arq, const PIPSocket::Address & rx_addr,
									 const PString  &vAgent, const endptr & ep)
									: m_requestSeqNumber(seqNum), m_vAgent(vAgent), m_reqEP (ep), m_ARQ(obj_arq), m_rx_addr(rx_addr)
				{
					m_enterTime     =  PTime() ;
					mm_ARQ          = m_ARQ ; 
					m_callReference = mm_ARQ.m_callReferenceValue ;
					m_epId          = mm_ARQ.m_endpointIdentifier.GetValue();
				};
			};

		private:
			
			list<WaitingARQ *> wArqList;
			int							   waitingTTL;
			int							   sequenceNumber;
			PMutex						 usedLock;
			PString						 m_CallCenterQueue ;			 // Virtual Queue ( sender name ).

	public:
		typedef std::list<WaitingARQ *>::iterator iterator;
		typedef std::list<WaitingARQ *>::const_iterator const_iterator;


		WaitingARQlist (int ttl,  const PString  &ccQueue ) : waitingTTL(ttl), sequenceNumber (1) ,m_CallCenterQueue (ccQueue)   {}
		~WaitingARQlist();

		bool InsertWaitingARQ(const H225_RasMessage & obj_arq, const PIPSocket::Address & rx_addr, const PString &ccQueue, const endptr & reqEP);
		void CheckWaitingARQ();
		
		PString  getVirtualQueue() const;
		
		bool     RouteToAlias (PString TargetAlias, PString SourceEpId, PString CallRef);
		bool     RouteReject (PString SourceEpId, PString CallRef);

		iterator FindByRequestSequenceNumber (int);
		iterator FindByEndpointIdAndCallRef (PString epId , int callRef) ;
		
		iterator FindByEndpointId (PString epid);
		iterator FindByCallRefNumber (int);
};


inline PString WaitingARQlist::WaitingARQ::getEndPointIdentifier () const {
	 return (m_epId )  ;
}

inline PString WaitingARQlist::WaitingARQ::getAgentQueue() const {
	return (m_vAgent);
}


inline int WaitingARQlist::WaitingARQ::DoARQ(void) const
{
	RasThread->DoARQ(m_rx_addr, m_ARQ);
	return 0;	// avoid VC++ 6.0 STL bug, must return something to be usable in mem_fun
}


inline int WaitingARQlist::WaitingARQ::DoARJ(void) const
{
	RasThread->ReplyARQ (m_reqEP, endptr(NULL), m_ARQ);
	return 0;	// avoid VC++ 6.0 STL bug, must return something to be usable in mem_fun
}


inline bool WaitingARQlist::WaitingARQ::CompReqId(int req_ID) const
{
		return (m_requestSeqNumber == req_ID);
}

inline bool WaitingARQlist::WaitingARQ::CompCallRef(int callRef) const
{
		return (m_callReference == callRef);
}

inline bool WaitingARQlist::WaitingARQ::CompEndpointId(PString ep) const
{
		return (m_epId == ep);
}


inline PString WaitingARQlist::getVirtualQueue() const {
	return (m_CallCenterQueue) ;
}

inline bool WaitingARQlist::WaitingARQ::IsTimeOut(int sec) const
{
		return (PTime() - m_enterTime) > sec*1000;
}

inline WaitingARQlist::~WaitingARQlist()
{
		DeleteObjectsInContainer<WaitingARQ>(wArqList);
}

inline WaitingARQlist::iterator WaitingARQlist::FindByRequestSequenceNumber (int reqId)
{
		return find_if(wArqList.begin(), wArqList.end(), bind2nd(mem_fun(&WaitingARQ::CompReqId), reqId ));
}

inline WaitingARQlist::iterator WaitingARQlist::FindByCallRefNumber (int callRef)
{

		return find_if(wArqList.begin(), wArqList.end(), bind2nd(mem_fun(&WaitingARQ::CompCallRef), callRef  ));
}

inline WaitingARQlist::iterator WaitingARQlist::FindByEndpointId (PString epId)
{
		return find_if(wArqList.begin(), wArqList.end(), bind2nd(mem_fun(&WaitingARQ::CompEndpointId), epId  ));
}

inline WaitingARQlist::iterator WaitingARQlist::FindByEndpointIdAndCallRef (PString epId , int callRef)
{
	iterator iter = wArqList.begin();
	while (iter != wArqList.end()) {
		bool ep = (*iter)->m_epId == epId ;
		bool rf = (*iter)-> m_callReference==callRef ;
		if ( ep && rf ) { 
			return iter;
		}	
		++iter;
	}
	return iter ;
	
}


#endif // HAS_WAITARQ
#endif // _WaitingARQ_h__
