/*
 * radauth.cxx
 *
 * RADIUS protocol authenticator module for GNU Gatekeeper. 
 * Please see docs/radauth.txt for more details.
 *
 * Copyright (c) 2003, Quarcom FHU, Michal Zygmuntowicz
 *
 * This work is published under the GNU Public License (GPL)
 * see file COPYING for details.
 * We also explicitely grant the right to link this code
 * with the OpenH323 library.
 *
 * $Log: radauth.cxx,v $
 * Revision 1.1.2.7  2003/05/29 17:21:22  zvision
 * Fixed compilation errors with OpenH323 versions prior to 1.11.5 (no H235AuthCAT)
 *
 * Revision 1.1.2.6  2003/05/28 13:25:19  zvision
 * Added alias based authentication (RadAliasAuth)
 *
 * Revision 1.1.2.5  2003/05/27 00:13:05  zvision
 * Smart Calling and Called -Station-Id selection (dialedDigits and partyNumber alias types preferred)
 *
 * Revision 1.1.2.4  2003/05/26 23:09:59  zvision
 * Added new OnSend and OnReceive hooks. LocalInterface config parameter introduced.
 *
 * Revision 1.1.2.3  2003/05/13 17:49:49  zvision
 * Removed acctPort. New includeFramedIP feature. Better tracing. Bug-fixes
 *
 * Revision 1.1.2.2  2003/04/29 14:56:27  zvision
 * Added H.235 capability matching
 *
 * Revision 1.1.2.1  2003/04/23 20:15:37  zvision
 * Initial revision
 *
 */
#include <ptlib.h>
#include <h225ras.h>
#include <h323pdu.h>
#include "gkauth.h"
#include "Toolkit.h"
#include "RasTbl.h"
#include "h323util.h"
#include "radauth.h"

// Settings for H.235 based module will be stored inside [RadAuth] config section
#define RadAuthConfigSectionName "RadAuth"
// Settings for alias based module will be stored inside [RadAliasAuth] config section
#define RadAliasAuthConfigSectionName "RadAliasAuth"


// OID for CAT (Cisco Access Token) algorithm
PString RadAuth::OID_CAT( "1.2.840.113548.10.1.2.1" );
// append RADIUS H.235 authenticator to the global list of authenticators
static GkAuthInit<RadAuth> RAD_A("RadAuth");
// append RADIUS Alias authenticator to the global list of authenticators
static GkAuthInit<RadAliasAuth> RAD_A_A("RadAliasAuth");


static PString GetConferenceIDString( const H225_ConferenceIdentifier& id )
{
	if( id.GetSize() < 16 )
		return PString::Empty();
		
	PString h323ConfId;
					
	for( int j = 0, i = 0; j < 4; j++ )
	{
		unsigned hex = (unsigned)(id[i++])<<24;
		hex |= (unsigned)(id[i++])<<16;
		hex |= (unsigned)(id[i++])<<8;
		hex |= (unsigned)(id[i++]);
							
		h323ConfId += PString( PString::Unsigned, (long)hex, 16 );
		if( j < 3 )
			h323ConfId += ' ';
	}

	return h323ConfId;
}

RadAuth::RadAuth( 
	PConfig* cfg, 
	const char* authName 
	)
	:
	GkAuthenticator( cfg, authName ),
	radiusServers( cfg->GetString(
			RadAuthConfigSectionName,"Servers",""
			).Tokenise( ";, |\t", FALSE ) 
		),
	sharedSecret( cfg->GetString(
			RadAuthConfigSectionName,"SharedSecret",""
			)
		),
	authPort( (WORD)(cfg->GetInteger(
			RadAuthConfigSectionName,"DefaultAuthPort"
			))
		),
	portBase( 1024 ),
	portMax( 65535 ),
	requestTimeout( cfg->GetInteger(
			RadAuthConfigSectionName,"RequestTimeout"
			)
		),
	idCacheTimeout( cfg->GetInteger(
			RadAuthConfigSectionName,"IdCacheTimeout"
			)
		),
	socketDeleteTimeout( cfg->GetInteger(
			RadAuthConfigSectionName,"SocketDeleteTimeout"
			)
		),
	numRequestRetransmissions( cfg->GetInteger(
			RadAuthConfigSectionName,"RequestRetransmissions"
			)
		),
	roundRobin( cfg->GetBoolean(
			RadAuthConfigSectionName,"RoundRobinServers", TRUE
			)
		),
	appendCiscoAttributes( cfg->GetBoolean(
			RadAuthConfigSectionName,"AppendCiscoAttributes", TRUE
			)
		),
	includeFramedIp( cfg->GetBoolean(
			RadAuthConfigSectionName, "IncludeEndpointIP", TRUE
			)
		),
	localInterface( cfg->GetString(
			RadAuthConfigSectionName, "LocalInterface", ""
			)
		),
	radiusClient( NULL )
{
	if( radiusServers.GetSize() < 1 )
	{
		PTRACE(1,"RADAUTH\tCannot build RADIUS authenticator"
			" - no RADIUS servers specified in the config"
			);
		return;
	}

	if( (!localInterface.IsEmpty()) && (!PIPSocket::IsLocalHost(localInterface)) )
	{
		PTRACE(1,"RADAUTH\tSpecified local interface - "<<localInterface
			<<" - does not belong to this machine"
			);
		localInterface = PString::Empty();
	}
	/// build RADIUS client
	radiusClient = new RadiusClient( 
		radiusServers[0],
		(radiusServers.GetSize() > 1) ? radiusServers[1] : PString::Empty(),
		localInterface
		);

	/// if there were specified more than two RADIUS servers, append them
	for( int i = 2; i < radiusServers.GetSize(); i++ )
		radiusClient->AppendServer( radiusServers[i] );	
		
	radiusClient->SetSharedSecret( sharedSecret );
	radiusClient->SetRoundRobinServers( roundRobin );
		
	if( authPort > 0 )
		radiusClient->SetAuthPort( authPort );
		
	if( requestTimeout > 0 )
		radiusClient->SetRequestTimeout( requestTimeout );
	if( idCacheTimeout > 0 )
		radiusClient->SetIdCacheTimeout( idCacheTimeout );
	if( socketDeleteTimeout > 0 )
		radiusClient->SetSocketDeleteTimeout( socketDeleteTimeout );
	if( numRequestRetransmissions > 0 )
		radiusClient->SetRetryCount( numRequestRetransmissions );
	
	PStringArray s = cfg->GetString(
		RadAuthConfigSectionName,"RadiusPortRange",""
		).Tokenise( "-" );

	// parse port range		
	if( s.GetSize() >= 2 )
	{ 
		unsigned p1 = s[0].AsUnsigned();
		unsigned p2 = s[1].AsUnsigned();
	
		// swap if base is greater than max
		if( p2 < p1 )
		{
			const unsigned temp = p1;
			p1 = p2;
			p2 = temp;
		}
		
		if( p1 > 65535 )
			p1 = 65535;
		if( p2 > 65535 )
			p2 = 65535;
	
		if( (p1 > 0) && (p2 > 0) )
		{
			portBase = (WORD)p1;
			portMax = (WORD)p2;
		}
	}
	
	radiusClient->SetClientPortRange( portBase, portMax-portBase+1 );
}

RadAuth::~RadAuth()
{
	delete radiusClient;
}

BOOL RadAuth::GetH235Capability(
	H225_ArrayOf_AuthenticationMechanism& mechanisms,
	H225_ArrayOf_PASN_ObjectId& algorithmOIDs
	) const
{
#ifdef OPENH323_NEWVERSION
	H235AuthCAT authenticator;
	authenticator.SetLocalId("dummy");
	authenticator.SetRemoteId("dummy");
	authenticator.SetPassword("dummy");
	return authenticator.SetCapability(mechanisms,algorithmOIDs);
#else
	return TRUE;
#endif
}

BOOL RadAuth::IsH235Capability(
	const H235_AuthenticationMechanism& mechanism,
	const PASN_ObjectId& algorithmOID
	) const
{
#ifdef OPENH323_NEWVERSION
	H235AuthCAT authenticator;
	return authenticator.IsCapability(mechanism,algorithmOID);
#else
	return FALSE;
#endif
}

BOOL RadAuth::CheckAliases( 
	const H225_ArrayOf_AliasAddress& aliases, 
	const PString& id 
	) const
{
	BOOL result = FALSE;
	
	for( PINDEX i = 0; i < aliases.GetSize(); i++ )
		if( H323GetAliasAddressString(aliases[i]) == id )
		{
			result = TRUE;
			break;
		}
	
	return result;
}

int RadAuth::Check(
	const H225_RegistrationRequest& rrq, 
	unsigned& rejectReason
	)
{
	if( radiusClient == NULL )
	{
		PTRACE(3,"RADAUTH\tRRQ Auth failed - NULL Radius client");
		if( defaultStatus == e_fail )
			rejectReason = H225_RegistrationRejectReason::e_undefinedReason;
		return defaultStatus;
	}

	// RRQ has to carry at least one terminalAlias
	if( !rrq.HasOptionalField(H225_RegistrationRequest::e_terminalAlias) )
	{
		if( defaultStatus == e_fail )
			rejectReason = H225_RegistrationRejectReason::e_securityDenial;
		return defaultStatus;
	}
		
	const H225_ArrayOf_AliasAddress& aliases = rrq.m_terminalAlias;
	
	// check for ClearTokens (CAT uses ClearTokens)
	if( rrq.HasOptionalField(H225_RegistrationRequest::e_tokens) )
	{
		const H225_ArrayOf_ClearToken& tokens = rrq.m_tokens;
		BOOL foundCAT = FALSE;
		
		// scan ClearTokens and find CATs
		for( PINDEX i = 0; i < tokens.GetSize(); i++ )
		{
			const H235_ClearToken& token = tokens[i];
			
			// is it CAT?
			if( token.m_tokenOID == OID_CAT )
			{
				foundCAT = TRUE;
				
				// these field are required for CAT
			  	if( !(token.HasOptionalField(H235_ClearToken::e_generalID)
					&& token.HasOptionalField(H235_ClearToken::e_random)
					&& token.HasOptionalField(H235_ClearToken::e_timeStamp)
					&& token.HasOptionalField(H235_ClearToken::e_challenge)) )
				{
					PTRACE(4,"RADAUTH\tRRQ Auth failed - CAT without all required fields");
					rejectReason = H225_RegistrationRejectReason::e_securityDenial;
					return e_fail;
				}
				
				// generalID should be present in the list of terminal aliases
				const PString id = token.m_generalID;
				if( !CheckAliases(aliases,id) )
				{
					PTRACE(4,"RADAUTH\tRRQ Auth failed - CAT m_generalID is not a valid alias");
					rejectReason = H225_RegistrationRejectReason::e_invalidTerminalAliases;
					return e_fail;
				}
					
				// CAT pseudo-random has to be one byte only
				const int randomInt = token.m_random;
					
				if( (randomInt < -127) || (randomInt > 255) )
				{
					PTRACE(4,"RADAUTH\tRRQ Auth failed - CAT m_random out of range");
					rejectReason = H225_RegistrationRejectReason::e_securityDenial;
					return e_fail;
				}
					
				// CAT challenge has to be 16 bytes
				if( token.m_challenge.GetValue().GetSize() < 16 )
				{
					PTRACE(4,"RADAUTH\tRRQ Auth failed - m_challenge less than 16 bytes");
					rejectReason = H225_RegistrationRejectReason::e_securityDenial;
					return e_fail;
				}
					
				// build RADIUS Access-Request
				RadiusPDU* pdu = radiusClient->BuildPDU();
				if( pdu == NULL )
				{
					PTRACE(3,"RADAUTH\tRRQ auth failed - could not to create Access-Request PDU");
					rejectReason = H225_RegistrationRejectReason::e_undefinedReason;
					continue;
				}

				pdu->SetCode( RadiusPDU::AccessRequest );
				
				// append User-Name
				*pdu += new RadiusAttr( RadiusAttr::UserName, id );
				
				// build CHAP-Password
				char password[17];
				password[0] = (BYTE)randomInt;
				memcpy(password+1,(const BYTE*)(token.m_challenge),16);
				
				*pdu += new RadiusAttr( RadiusAttr::ChapPassword,
					password,
					sizeof(password)
					);
				
				// Gk works as NAS point, so append NAS IP
				*pdu += new RadiusAttr( RadiusAttr::NasIpAddress, 
					localInterface.IsEmpty() 
						? Toolkit::Instance()->GetRouteTable()->GetLocalAddress()
						: PIPSocket::Address(localInterface)
					);
				// NAS-Identifier as Gk name
				*pdu += new RadiusAttr( RadiusAttr::NasIdentifier,
					Toolkit::Instance()->GKName()
					);
				// Gk does not have a concept of physical ports,
				// so define port type as NAS-Port-Virtual
				*pdu += new RadiusAttr( RadiusAttr::NasPortType, 
					RadiusAttr::NasPort_Virtual 
					);
				// RRQ service type is Login-User
				*pdu += new RadiusAttr( RadiusAttr::ServiceType,
					RadiusAttr::ST_Login
					);
					
				*pdu += new RadiusAttr( RadiusAttr::ChapChallenge,
					(int)(DWORD)token.m_timeStamp
					);
					
				if( includeFramedIp )
				{
					PIPSocket::Address addr;
				
					if( rrq.m_callSignalAddress.GetSize() > 0 )
					{
						if( GetIPFromTransportAddr(rrq.m_callSignalAddress[0],addr)
							&& addr.IsValid() )
							*pdu += new RadiusAttr( 
								RadiusAttr::FramedIpAddress,
								addr
								);
					}
					else if( rrq.m_rasAddress.GetSize() > 0 )
					{
						if( GetIPFromTransportAddr(rrq.m_rasAddress[0],addr) 
							&& addr.IsValid() )
							*pdu += new RadiusAttr( 
								RadiusAttr::FramedIpAddress,
								addr
								);
					}
				}
				
				// send request and wait for response
				RadiusPDU* response = NULL;
				BOOL result = OnSendPDU(*pdu,rrq,rejectReason)
					&& radiusClient->MakeRequest( *pdu, response ) 
					&& (response != NULL);
			
				delete pdu;
			
				if( !result )
				{
					delete response;
					return defaultStatus;
				}
				
				result = (response->GetCode() == RadiusPDU::AccessAccept);
				if( result )
					result = OnReceivedPDU(*response,rrq,rejectReason);
				else
					rejectReason = H225_RegistrationRejectReason::e_securityDenial;
					
				delete response;
				return result ? e_ok : e_fail;
			}
		}
		
		if( !foundCAT )
			if( defaultStatus == e_fail )
				PTRACE(3,"RADAUTH\tRRQ Auth failed - m_tokens without CAT");
			else if( defaultStatus != e_ok )
				PTRACE(4,"RADAUTH\tRRQ Auth undetermined - m_tokens without CAT");
	}	
	else
		if( defaultStatus == e_fail )
			PTRACE(3,"RADAUTH\tRRQ Auth failed - no m_tokens");
		else if( defaultStatus != e_ok )
			PTRACE(4,"RADAUTH\tRRQ Auth undetermined - no m_tokens");

	if( defaultStatus == e_fail )
		rejectReason = H225_RegistrationRejectReason::e_securityDenial;
		
	return defaultStatus;
}

int RadAuth::Check(
	const H225_AdmissionRequest& arq, 
	unsigned& rejectReason
	)
{
	if( radiusClient == NULL )
	{
		PTRACE(3,"RADAUTH\tARQ Auth failed - NULL Radius client");
		if( defaultStatus == e_fail )
			rejectReason = H225_AdmissionRejectReason::e_undefinedReason;
		return defaultStatus;
	}
	
	// check for ClearTokens
	if( arq.HasOptionalField(H225_AdmissionRequest::e_tokens) )
	{
		const H225_ArrayOf_ClearToken& tokens = arq.m_tokens;
		BOOL foundCAT = FALSE;
		
		// scan ClearTokens for CATs
		for( PINDEX i = 0; i < tokens.GetSize(); i++ )
		{
			const H235_ClearToken& token = tokens[i];
				
			// is it CAT?
			if( token.m_tokenOID == OID_CAT )
			{
				foundCAT = TRUE;
				
				// these field are required for CAT
			  	if( !(token.HasOptionalField(H235_ClearToken::e_generalID)
					&& token.HasOptionalField(H235_ClearToken::e_random)
					&& token.HasOptionalField(H235_ClearToken::e_timeStamp)
					&& token.HasOptionalField(H235_ClearToken::e_challenge)) )
				{
					PTRACE(4,"RADAUTH\tARQ Auth failed - CAT without all required fields");
					rejectReason = H225_AdmissionRejectReason::e_securityDenial;
					return e_fail;
				}
			
				const PString id = token.m_generalID;
					
				// CAT random has to be one byte only				
				const int randomInt = token.m_random;
					
				if( (randomInt < -127) || (randomInt > 255) )
				{
					PTRACE(4,"RADAUTH\tARQ Auth failed - CAT m_random out of range");
					rejectReason = H225_AdmissionRejectReason::e_securityDenial;
					return e_fail;
				}
					
				// CAT challenge has to be 16 bytes
				if( token.m_challenge.GetValue().GetSize() < 16 )
				{
					PTRACE(4,"RADAUTH\tARQ Auth failed - CAT m_challenge less than 16 bytes");
					rejectReason = H225_AdmissionRejectReason::e_securityDenial;
					return e_fail;
				}
					
				// build RADIUS Access-Request packet
				RadiusPDU* pdu = radiusClient->BuildPDU();
				if( pdu == NULL )
				{
					PTRACE(3,"RADAUTH\tARQ auth failed - could not to create Access-Request PDU");
					continue;
				}

				pdu->SetCode( RadiusPDU::AccessRequest );
				
				*pdu += new RadiusAttr( RadiusAttr::UserName, id );
				
				// build CHAP-Password
				char password[17];
				password[0] = (BYTE)randomInt;
				memcpy(password+1,(const BYTE*)(token.m_challenge),16);
				
				*pdu += new RadiusAttr( RadiusAttr::ChapPassword,
					password,
					sizeof(password)
					);
				
				// append CHAP-Challenge 
				*pdu += new RadiusAttr( RadiusAttr::ChapChallenge,
					(int)(DWORD)token.m_timeStamp
					);
				
				// Gk acts as NAS, so include NAS IP
				*pdu += new RadiusAttr( RadiusAttr::NasIpAddress, 
					localInterface.IsEmpty()
						? Toolkit::Instance()->GetRouteTable()->GetLocalAddress()
						: PIPSocket::Address(localInterface)
					);
					
				// NAS-Identifier as Gk name
				*pdu += new RadiusAttr( RadiusAttr::NasIdentifier,
					Toolkit::Instance()->GKName()
					);
					
				// NAS-Port-Type as Virtual, since Gk does
				// not care about physical ports concept
				*pdu += new RadiusAttr( RadiusAttr::NasPortType, 
					RadiusAttr::NasPort_Virtual 
					);
					
				// Service-Type is Login-User if originating the call
				// and Call Check if answering the call
				*pdu += new RadiusAttr( RadiusAttr::ServiceType,
					arq.m_answerCall ? 
						RadiusAttr::ST_CallCheck : RadiusAttr::ST_Login
					);
				
				endptr ep = RegistrationTable::Instance()->FindByEndpointId(
					arq.m_endpointIdentifier
					);
					
				if( includeFramedIp )
				{
					PIPSocket::Address addr;
				
					if( arq.HasOptionalField(arq.e_srcCallSignalAddress) )
					{
						if( GetIPFromTransportAddr(arq.m_srcCallSignalAddress,addr) 
							&& addr.IsValid() )
							*pdu += new RadiusAttr( 
								RadiusAttr::FramedIpAddress,
								addr
								);
					}
					else
					{
						if( ep 
							&& GetIPFromTransportAddr( ep->GetCallSignalAddress(), addr)
							&& addr.IsValid() )
							*pdu += new RadiusAttr( 
								RadiusAttr::FramedIpAddress,
								addr
								);
					}
				}
				
				// fill Calling-Station-Id and Called-Station-Id fields
				
				// Calling-Station-Id
				// Priority:
				//	ARQ.m_srcInfo[0] (first alias)
				//	generalID from CAT ClearToken (if the call originator)
					
				PString stationId;
				
				if( arq.m_srcInfo.GetSize() > 0 )
					stationId = GetBestAliasAddressString(
						arq.m_srcInfo,
						H225_AliasAddress::e_dialedDigits,
						H225_AliasAddress::e_partyNumber,
						H225_AliasAddress::e_h323_ID
						);
				
				if( stationId.IsEmpty() && ep && (!arq.m_answerCall) )
					stationId = GetBestAliasAddressString(
						ep->GetAliases(),
						H225_AliasAddress::e_dialedDigits,
						H225_AliasAddress::e_partyNumber,
						H225_AliasAddress::e_h323_ID
						);
					
				if( stationId.IsEmpty() && (!arq.m_answerCall) )
					stationId = id;
				
				if( !stationId.IsEmpty() )		
					*pdu += new RadiusAttr( 
						RadiusAttr::CallingStationId, 
						stationId			
						);
						
				stationId = PString::Empty();
					
				// Called-Station-Id
				// Priority:
				//	ARQ.m_destinationInfo[0] (first alias)
				//	ARQ.m_destCallSignalAddress::e_ipAddress
				//  generalID from CAT token (if answering the call)
				if( arq.HasOptionalField(H225_AdmissionRequest::e_destinationInfo) 
					&& (arq.m_destinationInfo.GetSize() > 0) )
					stationId = GetBestAliasAddressString(
						arq.m_destinationInfo,
						H225_AliasAddress::e_dialedDigits,
						H225_AliasAddress::e_partyNumber,
						H225_AliasAddress::e_h323_ID
						);
				
				if( stationId.IsEmpty() && ep && arq.m_answerCall )
					stationId = GetBestAliasAddressString(
						ep->GetAliases(),
						H225_AliasAddress::e_dialedDigits,
						H225_AliasAddress::e_partyNumber,
						H225_AliasAddress::e_h323_ID
						);
						
				if( stationId.IsEmpty() 
					&& arq.HasOptionalField(H225_AdmissionRequest::e_destCallSignalAddress) )
				{
					const H225_TransportAddress& tsap = arq.m_destCallSignalAddress;
					stationId = AsDotString(tsap);
				}
				
				if( stationId.IsEmpty() && arq.m_answerCall )
					stationId = id;
					
				if( stationId.IsEmpty() )
				{
					delete pdu;
					PTRACE(3,"RADAUTH\tARQ Auth failed - neither m_destinationInfo"
						" nor m_destCallSignalAddress found"
						);
					return e_fail; 	// ARQ without one of the above fields
									// should be rejected
				}
				
				*pdu += new RadiusAttr( RadiusAttr::CalledStationId,
					stationId
					);
			
				if( appendCiscoAttributes )
				{
					*pdu += new RadiusAttr(
						PString("h323-conf-id=") 
							+ GetConferenceIDString(arq.m_conferenceID),
						9, // VendorId = Cisco
						24 // Connection-ID (h323-conf-id)
						);
					
					*pdu += new RadiusAttr(
						PString(arq.m_answerCall 
							? "h323-call-origin=answer" 
							: "h323-call-origin=originate"),
						9, // VendorId = Cisco
						26 // Call-Origin (h323-call-origin)
						);
					*pdu += new RadiusAttr(
						PString("h323-call-type=VoIP"),
						9, // VendorId = Cisco
						27 // Call-Type (h323-call-type)
						);
					*pdu += new RadiusAttr(
						PString("h323-gw-id=" + Toolkit::Instance()->GKName()),
						9, // VendorId = Cisco
						33 // Call-Type (h323-gw-id)
						);
				}
					
				// send the request and wait for a response
				RadiusPDU* response = NULL;
				BOOL result = OnSendPDU(*pdu,arq,rejectReason) 
					&& radiusClient->MakeRequest( *pdu, response ) 
					&& (response != NULL);
			
				delete pdu;
			
				if( !result )
				{
					delete response;
					return defaultStatus;
				}
				
				// authenticated?
				result = (response->GetCode() == RadiusPDU::AccessAccept);
				if( result )
					result = OnReceivedPDU(*response,arq,rejectReason);
				else
					rejectReason = H225_AdmissionRejectReason::e_securityDenial;
					
				delete response;
				return result ? e_ok : e_fail;
			}
		}
		
		if( !foundCAT )
			if( defaultStatus == e_fail )
				PTRACE(3,"RADAUTH\tARQ Auth failed - m_tokens without CAT");
			else if( defaultStatus != e_ok )
				PTRACE(4,"RADAUTH\tARQ Auth undetermined - m_tokens without CAT");
	}	
	else
		if( defaultStatus == e_fail )
			PTRACE(3,"RADAUTH\tARQ Auth failed - no m_tokens");
		else if( defaultStatus != e_ok )
			PTRACE(4,"RADAUTH\tARQ Auth undetermined - no m_tokens");

	if( defaultStatus == e_fail )
		rejectReason = H225_AdmissionRejectReason::e_securityDenial;
	return defaultStatus;
}

BOOL RadAuth::OnSendPDU(
	RadiusPDU& pdu,
	const H225_RegistrationRequest& rrq,
	unsigned& rejectReason
	)
{
	return TRUE;
}

BOOL RadAuth::OnSendPDU(
	RadiusPDU& pdu,
	const H225_AdmissionRequest& rrq,
	unsigned& rejectReason
	)
{
	return TRUE;
}

BOOL RadAuth::OnReceivedPDU(
	RadiusPDU& pdu,
	const H225_RegistrationRequest& rrq,
	unsigned& rejectReason
	)
{
	return TRUE;
}

BOOL RadAuth::OnReceivedPDU(
	RadiusPDU& pdu,
	const H225_AdmissionRequest& rrq,
	unsigned& rejectReason
	)
{
	return TRUE;
}



RadAliasAuth::RadAliasAuth( 
	PConfig* cfg, 
	const char* authName 
	)
	:
	GkAuthenticator( cfg, authName ),
	radiusServers( cfg->GetString(
			RadAliasAuthConfigSectionName,"Servers",""
			).Tokenise( ";, |\t", FALSE ) 
		),
	sharedSecret( cfg->GetString(
			RadAliasAuthConfigSectionName,"SharedSecret",""
			)
		),
	authPort( (WORD)(cfg->GetInteger(
			RadAliasAuthConfigSectionName,"DefaultAuthPort"
			))
		),
	portBase( 1024 ),
	portMax( 65535 ),
	requestTimeout( cfg->GetInteger(
			RadAliasAuthConfigSectionName,"RequestTimeout"
			)
		),
	idCacheTimeout( cfg->GetInteger(
			RadAliasAuthConfigSectionName,"IdCacheTimeout"
			)
		),
	socketDeleteTimeout( cfg->GetInteger(
			RadAliasAuthConfigSectionName,"SocketDeleteTimeout"
			)
		),
	numRequestRetransmissions( cfg->GetInteger(
			RadAliasAuthConfigSectionName,"RequestRetransmissions"
			)
		),
	roundRobin( cfg->GetBoolean(
			RadAliasAuthConfigSectionName,"RoundRobinServers", TRUE
			)
		),
	appendCiscoAttributes( cfg->GetBoolean(
			RadAliasAuthConfigSectionName,"AppendCiscoAttributes", TRUE
			)
		),
	includeFramedIp( cfg->GetBoolean(
			RadAliasAuthConfigSectionName, "IncludeEndpointIP", TRUE
			)
		),
	localInterface( cfg->GetString(
			RadAliasAuthConfigSectionName, "LocalInterface", ""
			)
		),
	fixedUsername( cfg->GetString(
			RadAliasAuthConfigSectionName, "FixedUsername", ""
			)
		),
	fixedPassword( cfg->GetString(
			RadAliasAuthConfigSectionName, "FixedPassword", ""
			)
		),
	radiusClient( NULL )
{
	if( radiusServers.GetSize() < 1 )
	{
		PTRACE(1,"RADAUTH\tCannot build RADIUS Alias authenticator"
			" - no RADIUS servers specified in the config"
			);
		return;
	}

	if( (!localInterface.IsEmpty()) && (!PIPSocket::IsLocalHost(localInterface)) )
	{
		PTRACE(1,"RADAUTH\tSpecified local interface - "<<localInterface
			<<" - does not belong to this machine"
			);
		localInterface = PString::Empty();
	}
	/// build RADIUS client
	radiusClient = new RadiusClient( 
		radiusServers[0],
		(radiusServers.GetSize() > 1) ? radiusServers[1] : PString::Empty(),
		localInterface
		);

	/// if there were specified more than two RADIUS servers, append them
	for( int i = 2; i < radiusServers.GetSize(); i++ )
		radiusClient->AppendServer( radiusServers[i] );	
		
	radiusClient->SetSharedSecret( sharedSecret );
	radiusClient->SetRoundRobinServers( roundRobin );
		
	if( authPort > 0 )
		radiusClient->SetAuthPort( authPort );
		
	if( requestTimeout > 0 )
		radiusClient->SetRequestTimeout( requestTimeout );
	if( idCacheTimeout > 0 )
		radiusClient->SetIdCacheTimeout( idCacheTimeout );
	if( socketDeleteTimeout > 0 )
		radiusClient->SetSocketDeleteTimeout( socketDeleteTimeout );
	if( numRequestRetransmissions > 0 )
		radiusClient->SetRetryCount( numRequestRetransmissions );
	
	PStringArray s = cfg->GetString(
		RadAliasAuthConfigSectionName,"RadiusPortRange",""
		).Tokenise( "-" );

	// parse port range		
	if( s.GetSize() >= 2 )
	{ 
		unsigned p1 = s[0].AsUnsigned();
		unsigned p2 = s[1].AsUnsigned();
	
		// swap if base is greater than max
		if( p2 < p1 )
		{
			const unsigned temp = p1;
			p1 = p2;
			p2 = temp;
		}
		
		if( p1 > 65535 )
			p1 = 65535;
		if( p2 > 65535 )
			p2 = 65535;
	
		if( (p1 > 0) && (p2 > 0) )
		{
			portBase = (WORD)p1;
			portMax = (WORD)p2;
		}
	}
	
	radiusClient->SetClientPortRange( portBase, portMax-portBase+1 );
}

RadAliasAuth::~RadAliasAuth()
{
	delete radiusClient;
}

BOOL RadAliasAuth::CheckAliases( 
	const H225_ArrayOf_AliasAddress& aliases, 
	const PString& id 
	) const
{
	BOOL result = FALSE;
	
	for( PINDEX i = 0; i < aliases.GetSize(); i++ )
		if( H323GetAliasAddressString(aliases[i]) == id )
		{
			result = TRUE;
			break;
		}
	
	return result;
}

int RadAliasAuth::Check(
	const H225_RegistrationRequest& rrq, 
	unsigned& rejectReason
	)
{
	if( radiusClient == NULL )
	{
		PTRACE(3,"RADAUTH\tRRQ AliasAuth failed - NULL RADIUS Client");
		if( defaultStatus != e_ok )
			rejectReason = H225_ReleaseCompleteReason::e_undefinedReason;
		return defaultStatus;
	}

	// build RADIUS Access-Request
	RadiusPDU* pdu = radiusClient->BuildPDU();
	if( pdu == NULL )
	{
		if( defaultStatus == e_fail )
			PTRACE(3,"RADAUTH\tRRQ AliasAuth failed - could not to create Access-Request PDU");
		else if( defaultStatus != e_ok )
			PTRACE(4,"RADAUTH\tRRQ AliasAuth undetermined - could not to create Access-Request PDU");
		return defaultStatus;
	}

	pdu->SetCode( RadiusPDU::AccessRequest );


	PString username;				
	PIPSocket::Address addr;

	if( rrq.HasOptionalField(H225_RegistrationRequest::e_terminalAlias) 
		&& (rrq.m_terminalAlias.GetSize() > 0) )
		username = GetBestAliasAddressString(
			rrq.m_terminalAlias,
			H225_AliasAddress::e_h323_ID,
			H225_AliasAddress::e_dialedDigits
			);

	if( username.IsEmpty() && (rrq.m_callSignalAddress.GetSize() > 0) )
		if( GetIPFromTransportAddr(rrq.m_callSignalAddress[0],addr) 
			&& addr.IsValid() )
			username = addr.AsString();
	
	if( username.IsEmpty() && (rrq.m_rasAddress.GetSize() > 0) )
		if( GetIPFromTransportAddr(rrq.m_rasAddress[0],addr) 
			&& addr.IsValid() )
			username = addr.AsString();
							
	if( username.IsEmpty() 
		&& rrq.HasOptionalField(H225_RegistrationRequest::e_endpointIdentifier) )
		username = rrq.m_endpointIdentifier;

	if( username.IsEmpty() && fixedUsername.IsEmpty() )
	{
		PTRACE(2,"RADAUTH\tRRQ AliasAuth failed - neither FixedUsername"
			" nor alias inside RRQ were found"
			);
		rejectReason = H225_RegistrationRejectReason::e_securityDenial;
		return defaultStatus;
	}
	
	// append User-Name
    *pdu += new RadiusAttr( RadiusAttr::UserName, 
		fixedUsername.IsEmpty() ? username : fixedUsername 
		);
	*pdu += new RadiusAttr( RadiusAttr::UserPassword, 
		fixedPassword.IsEmpty() 
			? (fixedUsername.IsEmpty() ? username : fixedUsername)
			: fixedPassword  
			);
			
	// Gk works as NAS point, so append NAS IP
	*pdu += new RadiusAttr( RadiusAttr::NasIpAddress, 
		localInterface.IsEmpty() 
			? Toolkit::Instance()->GetRouteTable()->GetLocalAddress()
			: PIPSocket::Address(localInterface)
		);
	*pdu += new RadiusAttr( RadiusAttr::NasPortType, RadiusAttr::NasPort_Virtual );
	*pdu += new RadiusAttr( RadiusAttr::ServiceType, RadiusAttr::ST_Login );
	
	if( includeFramedIp )
	{
		if( rrq.m_callSignalAddress.GetSize() > 0 )
		{
			if( GetIPFromTransportAddr(rrq.m_callSignalAddress[0],addr)
				&& addr.IsValid() )
				*pdu += new RadiusAttr( 
					RadiusAttr::FramedIpAddress,
					addr
					);
		}
		else if( rrq.m_rasAddress.GetSize() > 0 )
		{
			if( GetIPFromTransportAddr(rrq.m_rasAddress[0],addr) 
				&& addr.IsValid() )
				*pdu += new RadiusAttr( 
					RadiusAttr::FramedIpAddress,
					addr
					);
		}
	}
					
	// send request and wait for response
	RadiusPDU* response = NULL;
	BOOL result = OnSendPDU(*pdu,rrq,rejectReason)
		&& radiusClient->MakeRequest( *pdu, response ) 
		&& (response != NULL);
			
	delete pdu;
			
	if( !result )
	{
		delete response;
		if( defaultStatus == e_fail )
			PTRACE(3,"RADAUTH\tRRQ AliasAuth failed - could not send Access-Request PDU");
		else if( defaultStatus != e_ok )
			PTRACE(4,"RADAUTH\tRRQ AliasAuth undetermined - could not send Access-Request PDU");
		return defaultStatus;
	}
				
	result = (response->GetCode() == RadiusPDU::AccessAccept);
	if( result )
		result = OnReceivedPDU(*response,rrq,rejectReason);
	else
		rejectReason = H225_RegistrationRejectReason::e_securityDenial;
					
	delete response;
	return result ? e_ok : e_fail;
}

int RadAliasAuth::Check(
	const H225_AdmissionRequest& arq, 
	unsigned& rejectReason
	)
{
	if( radiusClient == NULL )
	{
		PTRACE(3,"RADAUTH\tARQ AliasAuth failed - NULL Radius client");
		if( defaultStatus == e_fail )
			rejectReason = H225_AdmissionRejectReason::e_undefinedReason;
		return defaultStatus;
	}
	
	// build RADIUS Access-Request packet
	RadiusPDU* pdu = radiusClient->BuildPDU();
	if( pdu == NULL )
	{
		if( defaultStatus == e_fail )
			PTRACE(3,"RADAUTH\tARQ AliasAuth failed - could not to create Access-Request PDU");
		else if( defaultStatus != e_ok )
			PTRACE(4,"RADAUTH\tARQ AliasAuth undetermined - could not to create Access-Request PDU");
		return defaultStatus;
	}

	pdu->SetCode( RadiusPDU::AccessRequest );
				
	PString username;				
	PIPSocket::Address addr;

	endptr ep = RegistrationTable::Instance()->FindByEndpointId(
		arq.m_endpointIdentifier
		);
		
	if( ep && (ep->GetAliases().GetSize() > 0) )
		username = GetBestAliasAddressString(
			ep->GetAliases(),
			H225_AliasAddress::e_h323_ID,
			H225_AliasAddress::e_dialedDigits
			);

	if( username.IsEmpty() && (arq.m_srcInfo.GetSize() > 0) )
		username = GetBestAliasAddressString(
			arq.m_srcInfo,
			H225_AliasAddress::e_h323_ID,
			H225_AliasAddress::e_dialedDigits
			);
							
	if( username.IsEmpty() 
		&& arq.HasOptionalField(H225_AdmissionRequest::e_srcCallSignalAddress) )
	{
		if( GetIPFromTransportAddr(arq.m_srcCallSignalAddress,addr) 
			&& addr.IsValid() )
			username = addr.AsString();
	}
	
	if( username.IsEmpty() && ep )
	{
		if( GetIPFromTransportAddr(ep->GetCallSignalAddress(),addr) 
			&& addr.IsValid() )
			username = addr.AsString();
	}
	
	if( username.IsEmpty() )
		username = arq.m_endpointIdentifier;

	if( username.IsEmpty() && fixedUsername.IsEmpty() )
	{
		PTRACE(2,"RADAUTH\tARQ AliasAuth failed - neither FixedUsername"
			" nor alias inside ARQ were found"
			);
		rejectReason = H225_AdmissionRejectReason::e_securityDenial;
		return defaultStatus;
	}
	
	// append User-Name
    *pdu += new RadiusAttr( RadiusAttr::UserName, 
		fixedUsername.IsEmpty() ? username : fixedUsername 
		);
		
	*pdu += new RadiusAttr( RadiusAttr::UserPassword, 
		fixedPassword.IsEmpty() 
			? (fixedUsername.IsEmpty() ? username : fixedUsername)
			: fixedPassword  
			);
			
	// Gk works as NAS point, so append NAS IP
	*pdu += new RadiusAttr( RadiusAttr::NasIpAddress, 
		localInterface.IsEmpty() 
			? Toolkit::Instance()->GetRouteTable()->GetLocalAddress()
			: PIPSocket::Address(localInterface)
		);
	*pdu += new RadiusAttr( RadiusAttr::NasPortType, RadiusAttr::NasPort_Virtual );
	// Service-Type is Login-User if originating the call
	// and Call Check if answering the call
	*pdu += new RadiusAttr( RadiusAttr::ServiceType,
		arq.m_answerCall ? 
			RadiusAttr::ST_CallCheck : RadiusAttr::ST_Login
		);
				
	if( includeFramedIp )
	{
		if( arq.HasOptionalField(arq.e_srcCallSignalAddress) )
		{
			if( GetIPFromTransportAddr(arq.m_srcCallSignalAddress,addr) 
				&& addr.IsValid() )
				*pdu += new RadiusAttr( 
					RadiusAttr::FramedIpAddress,
					addr
					);
		}
		else
			if( ep && GetIPFromTransportAddr( ep->GetCallSignalAddress(), addr)
				&& addr.IsValid() )
				*pdu += new RadiusAttr( 
					RadiusAttr::FramedIpAddress,
					addr
					);
	}
				
	// fill Calling-Station-Id and Called-Station-Id fields
				
	// Calling-Station-Id
	// Priority:
	//	ARQ.m_srcInfo[0] (first alias)
	//	generalID from CAT ClearToken (if the call originator)
					
	PString stationId;
				
	if( arq.m_srcInfo.GetSize() > 0 )
		stationId = GetBestAliasAddressString(
			arq.m_srcInfo,
			H225_AliasAddress::e_dialedDigits,
			H225_AliasAddress::e_partyNumber,
			H225_AliasAddress::e_h323_ID
			);
			
	if( stationId.IsEmpty() && ep && (!arq.m_answerCall) )
		stationId = GetBestAliasAddressString(
			ep->GetAliases(),
			H225_AliasAddress::e_dialedDigits,
			H225_AliasAddress::e_partyNumber,
			H225_AliasAddress::e_h323_ID
			);
			
	if( stationId.IsEmpty() && (!arq.m_answerCall) )
		stationId = fixedUsername.IsEmpty() ? username : fixedUsername;
				
	if( !stationId.IsEmpty() )		
		*pdu += new RadiusAttr( 
			RadiusAttr::CallingStationId, 
			stationId			
			);
						
	stationId = PString::Empty();
				
	// Called-Station-Id
	// Priority:
	//	ARQ.m_destinationInfo[0] (first alias)
	//	ARQ.m_destCallSignalAddress::e_ipAddress
	//  generalID from CAT token (if answering the call)
	if( arq.HasOptionalField(H225_AdmissionRequest::e_destinationInfo) 
		&& (arq.m_destinationInfo.GetSize() > 0) )
		stationId = GetBestAliasAddressString(
			arq.m_destinationInfo,
			H225_AliasAddress::e_dialedDigits,
			H225_AliasAddress::e_partyNumber,
			H225_AliasAddress::e_h323_ID
			);
				
	if( stationId.IsEmpty() 
		&& arq.HasOptionalField(H225_AdmissionRequest::e_destCallSignalAddress) )
	{
		const H225_TransportAddress& tsap = arq.m_destCallSignalAddress;
		stationId = AsDotString(tsap);
	}
				
	if( stationId.IsEmpty() && arq.m_answerCall )
		stationId = fixedUsername.IsEmpty() ? username : fixedUsername;
					
	if( stationId.IsEmpty() )
	{
		delete pdu;
		PTRACE(3,"RADAUTH\tARQ AliasAuth failed - neither m_destinationInfo"
			" nor m_destCallSignalAddress found"
			);
		return e_fail; 	// ARQ without one of the above fields
						// should be rejected
	}
				
	*pdu += new RadiusAttr( RadiusAttr::CalledStationId, stationId );
			
	if( appendCiscoAttributes )
	{
		*pdu += new RadiusAttr(
			PString("h323-conf-id=") 
				+ GetConferenceIDString(arq.m_conferenceID),
			9, // VendorId = Cisco
			24 // Connection-ID (h323-conf-id)
			);
					
		*pdu += new RadiusAttr(
			PString(arq.m_answerCall 
				? "h323-call-origin=answer" 
				: "h323-call-origin=originate"),
			9, // VendorId = Cisco
			26 // Call-Origin (h323-call-origin)
			);
		*pdu += new RadiusAttr(
			PString("h323-call-type=VoIP"),
			9, // VendorId = Cisco
			27 // Call-Type (h323-call-type)
			);
		*pdu += new RadiusAttr(
			PString("h323-gw-id=" + Toolkit::Instance()->GKName()),
			9, // VendorId = Cisco
			33 // Call-Type (h323-gw-id)
			);
	}
					
	// send the request and wait for a response
	RadiusPDU* response = NULL;
	BOOL result = OnSendPDU(*pdu,arq,rejectReason) 
		&& radiusClient->MakeRequest( *pdu, response ) 
		&& (response != NULL);
			
	delete pdu;
			
	if( !result )
	{
		delete response;
		return defaultStatus;
	}
				
	// authenticated?
	result = (response->GetCode() == RadiusPDU::AccessAccept);
	if( result )
		result = OnReceivedPDU(*response,arq,rejectReason);
	else
		rejectReason = H225_AdmissionRejectReason::e_securityDenial;
					
	delete response;
	return result ? e_ok : e_fail;
}

BOOL RadAliasAuth::OnSendPDU(
	RadiusPDU& pdu,
	const H225_RegistrationRequest& rrq,
	unsigned& rejectReason
	)
{
	return TRUE;
}

BOOL RadAliasAuth::OnSendPDU(
	RadiusPDU& pdu,
	const H225_AdmissionRequest& rrq,
	unsigned& rejectReason
	)
{
	return TRUE;
}

BOOL RadAliasAuth::OnReceivedPDU(
	RadiusPDU& pdu,
	const H225_RegistrationRequest& rrq,
	unsigned& rejectReason
	)
{
	return TRUE;
}

BOOL RadAliasAuth::OnReceivedPDU(
	RadiusPDU& pdu,
	const H225_AdmissionRequest& rrq,
	unsigned& rejectReason
	)
{
	return TRUE;
}
