/*
 * radauth.h
 *
 * RADIUS protocol authenticator modules for GNU Gatekeeper. 
 * H.235 based and alias based authentication schemes are supported.
 * Please see docs/radauth.txt for more details.
 *
 * Copyright (c) 2003, Quarcom FHU, Michal Zygmuntowicz
 *
 * This work is published under the GNU Public License (GPL)
 * see file COPYING for details.
 * We also explicitely grant the right to link this code
 * with the OpenH323 library.
 *
 * $Log: radauth.h,v $
 * Revision 1.1.2.5  2003/05/28 13:25:19  zvision
 * Added alias based authentication (RadAliasAuth)
 *
 * Revision 1.1.2.4  2003/05/26 23:08:18  zvision
 * New OnSend and OnReceive hooks.
 * LocalInterface config parameter Introduced.
 *
 * Revision 1.1.2.3  2003/05/13 17:48:43  zvision
 * Removed acctPort. New includeFramedIP feature.
 *
 * Revision 1.1.2.2  2003/04/29 14:56:26  zvision
 * Added H.235 capability matching
 *
 * Revision 1.1.2.1  2003/04/23 20:16:25  zvision
 * Initial revision
 *
 */
#ifndef __RADAUTH_H
#define __RADAUTH_H

#include <ptlib.h>
#include <h225.h>
#include "gkauth.h"
#include "radproto.h"

/**
 * Gatekeeper authenticator module for RADIUS protocol.
 * Currently it supports user authentication through
 * CATs (Cisco Access Tokens) carried inside RRQ and ARQ
 * RAS messages. If your software does not support CATs,
 * please take a look at OpenH323 H235AuthCAT authenticator class
 * - it provides an implementation for CATs.
 */
class RadAuth : public GkAuthenticator 
{
public:
	/** Create GkAuthenticator for RADIUS protocol
	*/
	RadAuth( 
		/// authenticator settings
		PConfig* cfg, 
		/// authenticator name from Gatekeeper::Auth section
		const char* authName 
		);
		
	/// Destroy the authenticator
	virtual ~RadAuth();
	
	/** @return
		TRUE if this authenticator provides H.235 compatible security.
	*/
	virtual BOOL IsH235Capable() const { return TRUE; }
	
	/** If the authenticator supports H.235 security,
		this call returns H.235 security capabilities
		associated with it.
		
		@return
		TRUE is H.235 security is supported and capabilities
		has been set.
	*/
	virtual BOOL GetH235Capability(
		/// append supported authentication mechanism to this array
		H225_ArrayOf_AuthenticationMechanism& mechanisms,
		/// append supported algorithm OIDs for the given authentication
		/// mechanism
		H225_ArrayOf_PASN_ObjectId& algorithmOIDs
		) const;

	/** Check if this authenticator supports the given
		H.235 capability (mechanism+algorithmOID).
		
		@return
		TRUE if the capability is supported.
	*/
	virtual BOOL IsH235Capability(
		/// authentication mechanism
		const H235_AuthenticationMechanism& mechanism,
		/// algorithm OID for the given authentication mechanism
		const PASN_ObjectId& algorithmOID
		) const;

protected:		
	/** Hook for adding/modifying pdu before it is sent.
		It can be used to add custom attributes, for example.
		
		@return
		TRUE if PDU should be sent, FALSE to reject RRQ
		(rejectReason can be set to indicate a particular reason).
	*/
	virtual BOOL OnSendPDU(
		RadiusPDU& pdu, /// PDU to be sent
		const H225_RegistrationRequest& rrq, /// RRQ being processed
		unsigned& rejectReason /// reject reason on return FALSE
		);

	/** Hook for adding/modifying pdu before it is sent.
		It can be used to add custom attributes, for example.
		
		@return
		TRUE if PDU should be sent, FALSE to reject ARQ
		(rejectReason can be set to indicate a particular reason).
	*/
	virtual BOOL OnSendPDU(
		RadiusPDU& pdu, /// PDU to be sent
		const H225_AdmissionRequest& rrq, /// ARQ being processed
		unsigned& rejectReason /// reject reason on return FALSE
		);

	/** Hook for processing pdu after it is received.
		It can be used to process custom attributes, for example.
		
		@return
		TRUE if PDU should be accepted, FALSE to reject RRQ
		(rejectReason can be set to indicate a particular reason).
	*/
	virtual BOOL OnReceivedPDU(
		RadiusPDU& pdu, /// received PDU 
		const H225_RegistrationRequest& rrq, /// RRQ being processed
		unsigned& rejectReason /// reject reason on return FALSE
		);

	/** Hook for processing pdu after it is received.
		It can be used to process custom attributes, for example.
		
		@return
		TRUE if PDU should be accepted, FALSE to reject ARQ
		(rejectReason can be set to indicate a particular reason).
	*/
	virtual BOOL OnReceivedPDU(
		RadiusPDU& pdu, /// received PDU
		const H225_AdmissionRequest& rrq, /// ARQ being processed
		unsigned& rejectReason /// reject reason on return FALSE
		);
		
private:
	/** Scan the array of 'aliases' for 'id' alias.
	
		@return
		TRUE if 'id' is found in the 'aliases' array.
	*/
	virtual BOOL CheckAliases( 
		/// array of aliases to be searched
		const H225_ArrayOf_AliasAddress& aliases, 
		/// alias to be searched for
		const PString& id 
		) const;
	
	
	/** Authenticate using data from RRQ RAS message.
	
		@return:
		#GkAuthenticator::Status enum# with the result of authentication.
	*/
	virtual int Check(
		/// RRQ RAS message to be authenticated
		const H225_RegistrationRequest & rrq, 
		/// reference to the variable, that can be set 
		/// to custom H225_RegistrationRejectReason
		/// if the check fails
		unsigned& rejectReason
		);
		
	/** Authenticate using data from ARQ RAS message.
	
		@return:
		#GkAuthenticator::Status enum# with the result of authentication.
	*/
	virtual int Check(
		/// ARQ nessage to be authenticated
		const H225_AdmissionRequest &, 
		/// reference to the variable, that can be set 
		/// to custom H225_AdmissionRejectReason
		/// if the check fails
		unsigned& rejectReason
		); 
		
//	virtual int Check(const H225_UnregistrationRequest &, unsigned& );
//	virtual int Check(const H225_GatekeeperRequest &, unsigned& );
//	virtual int Check(const H225_BandwidthRequest &, unsigned& );
//	virtual int Check(const H225_DisengageRequest &, unsigned& );
//	virtual int Check(const H225_LocationRequest &, unsigned& );
//	virtual int Check(const H225_InfoRequest &, unsigned& );

	/* No copy constructor allowed */
	RadAuth( const RadAuth& );
	/* No operator= allowed */
	RadAuth& operator=( const RadAuth& );
	
private:
	/// array of configured RADIUS server names
	PStringArray radiusServers;
	/// shared secret for gk client<->RADIUS server authorization
	PString sharedSecret;
	/// default port that will be used for sending RADIUS auth
	/// requests
	WORD authPort;
	/// base port number for UDP client socket allocation
	WORD portBase;
	/// max port number for UDP client socket allocation
	WORD portMax;
	/// timeout (ms) for a single RADIUS request
	unsigned requestTimeout;
	/// timeout (ms) for RADIUS requests IDs to be unique
	unsigned idCacheTimeout;
	/// timeout (ms) for unused sockets to be deleted
	unsigned socketDeleteTimeout;
	/// how many times to transmit a single request (1==no retransmission)
	/// to a single RADIUS server
	unsigned numRequestRetransmissions;
	/// retransmission fashion: 
	/// 	FALSE - do #numRequestRetransmissions# for server A,
	///				then do #numRequestRetransmissions# for server B, etc.
	///		TRUE - transmit request to server A, then to server B, etc.
	///				the whole procedure repeat #numRequestRetransmissions# times
	BOOL roundRobin;
	/// if TRUE Cisco VSAs are appended to the RADIUS packets
	BOOL appendCiscoAttributes;
	/// if TRUE endpoint IP is placed inside Framed-IP-Address attribute
	BOOL includeFramedIp;
	/// Local interface RADIUS client should be bound to (multihomed hosts)
	PString localInterface;	
	/// RADIUS protocol client class associated with this authenticator
	RadiusClient* radiusClient;
	/// OID (Object Identifier) for CAT alghoritm
	static PString OID_CAT;
};

/** RADIUS Alias Authentication module.
	It authenticates endpoints/calls using non-H.235 
	attributes (alias,address,etc).
*/
class RadAliasAuth : public GkAuthenticator 
{
public:
	/** Create GkAuthenticator for RADIUS Alias authenticator
	*/
	RadAliasAuth( 
		/// authenticator settings
		PConfig* cfg, 
		/// authenticator name from Gatekeeper::Auth section
		const char* authName 
		);
		
	/// Destroy the authenticator
	virtual ~RadAliasAuth();
	
	/** @return
		FALSE - this authenticator does not support H.235.
	*/
	virtual BOOL IsH235Capable() const { return FALSE; }

protected:		
	/** Hook for adding/modifying pdu before it is sent.
		It can be used to add custom attributes, for example.
		
		@return
		TRUE if PDU should be sent, FALSE to reject RRQ
		(rejectReason can be set to indicate a particular reason).
	*/
	virtual BOOL OnSendPDU(
		RadiusPDU& pdu, /// PDU to be sent
		const H225_RegistrationRequest& rrq, /// RRQ being processed
		unsigned& rejectReason /// reject reason on return FALSE
		);

	/** Hook for adding/modifying pdu before it is sent.
		It can be used to add custom attributes, for example.
		
		@return
		TRUE if PDU should be sent, FALSE to reject ARQ
		(rejectReason can be set to indicate a particular reason).
	*/
	virtual BOOL OnSendPDU(
		RadiusPDU& pdu, /// PDU to be sent
		const H225_AdmissionRequest& rrq, /// ARQ being processed
		unsigned& rejectReason /// reject reason on return FALSE
		);

	/** Hook for processing pdu after it is received.
		It can be used to process custom attributes, for example.
		
		@return
		TRUE if PDU should be accepted, FALSE to reject RRQ
		(rejectReason can be set to indicate a particular reason).
	*/
	virtual BOOL OnReceivedPDU(
		RadiusPDU& pdu, /// received PDU 
		const H225_RegistrationRequest& rrq, /// RRQ being processed
		unsigned& rejectReason /// reject reason on return FALSE
		);

	/** Hook for processing pdu after it is received.
		It can be used to process custom attributes, for example.
		
		@return
		TRUE if PDU should be accepted, FALSE to reject ARQ
		(rejectReason can be set to indicate a particular reason).
	*/
	virtual BOOL OnReceivedPDU(
		RadiusPDU& pdu, /// received PDU
		const H225_AdmissionRequest& rrq, /// ARQ being processed
		unsigned& rejectReason /// reject reason on return FALSE
		);
		
private:
	/** Scan the array of 'aliases' for 'id' alias.
	
		@return
		TRUE if 'id' is found in the 'aliases' array.
	*/
	virtual BOOL CheckAliases( 
		/// array of aliases to be searched
		const H225_ArrayOf_AliasAddress& aliases, 
		/// alias to be searched for
		const PString& id 
		) const;
	
	
	/** Authenticate using data from RRQ RAS message.
	
		@return:
		#GkAuthenticator::Status enum# with the result of authentication.
	*/
	virtual int Check(
		/// RRQ RAS message to be authenticated
		const H225_RegistrationRequest & rrq, 
		/// reference to the variable, that can be set 
		/// to custom H225_RegistrationRejectReason
		/// if the check fails
		unsigned& rejectReason
		);
		
	/** Authenticate using data from ARQ RAS message.
	
		@return:
		#GkAuthenticator::Status enum# with the result of authentication.
	*/
	virtual int Check(
		/// ARQ nessage to be authenticated
		const H225_AdmissionRequest &, 
		/// reference to the variable, that can be set 
		/// to custom H225_AdmissionRejectReason
		/// if the check fails
		unsigned& rejectReason
		); 
		
	/* No copy constructor allowed */
	RadAliasAuth( const RadAliasAuth& );
	/* No operator= allowed */
	RadAliasAuth& operator=( const RadAliasAuth& );
	
private:
	/// array of configured RADIUS server names
	PStringArray radiusServers;
	/// shared secret for gk client<->RADIUS server authorization
	PString sharedSecret;
	/// default port that will be used for sending RADIUS auth
	/// requests
	WORD authPort;
	/// base port number for UDP client socket allocation
	WORD portBase;
	/// max port number for UDP client socket allocation
	WORD portMax;
	/// timeout (ms) for a single RADIUS request
	unsigned requestTimeout;
	/// timeout (ms) for RADIUS requests IDs to be unique
	unsigned idCacheTimeout;
	/// timeout (ms) for unused sockets to be deleted
	unsigned socketDeleteTimeout;
	/// how many times to transmit a single request (1==no retransmission)
	/// to a single RADIUS server
	unsigned numRequestRetransmissions;
	/// retransmission fashion: 
	/// 	FALSE - do #numRequestRetransmissions# for server A,
	///				then do #numRequestRetransmissions# for server B, etc.
	///		TRUE - transmit request to server A, then to server B, etc.
	///				the whole procedure repeat #numRequestRetransmissions# times
	BOOL roundRobin;
	/// if TRUE Cisco VSAs are appended to the RADIUS packets
	BOOL appendCiscoAttributes;
	/// if TRUE endpoint IP is placed inside Framed-IP-Address attribute
	BOOL includeFramedIp;
	/// local interface RADIUS client should be bound to (multihomed hosts)
	PString localInterface;	
	/// fixed username to be send in RADIUS requests instead of alias
	PString fixedUsername;
	/// fixed password to be send in RADIUS requests instead of duplicating User-Name
	PString fixedPassword;
	/// RADIUS protocol client class associated with this authenticator
	RadiusClient* radiusClient;
};

#endif /* __RADAUTH_H */
